import httpx

from .core.authorizations import ClientInfoFactory
from .core.client import FlyMyAI, AsyncFlyMyAI
from .core.exceptions import FlyMyAIPredictException, FlyMyAIExceptionGroup


__all__ = [
    "run",
    "httpx",
    "async_run",
    "FlyMyAI",
    "AsyncFlyMyAI",
    "ClientInfoFactory",
    "FlyMyAIExceptionGroup",
    "FlyMyAIPredictException",
]


client = FlyMyAI
async_client = AsyncFlyMyAI
run = client.run_predict
async_run = async_client.arun_predict



import io
import mimetypes
import pathlib
import uuid
from io import BytesIO
from typing import Union, BinaryIO, Any

from .base_field import BaseField

_BinaryInput = Union[bytes, pathlib.Path, BinaryIO, str]
_IOOutput = Union[BinaryIO, io.BytesIO]
_FieldOutput = tuple[str, _IOOutput, str]  # filename, io[binary], mime


def is_binary_input(value: _BinaryInput) -> bool:
    if isinstance(value, (bytes, BinaryIO)):
        return True
    if isinstance(value, str):
        try:
            pathlib.Path(value).resolve()
            return True
        except FileNotFoundError:
            return False
    if isinstance(value, pathlib.Path):
        return True
    return False


class BinaryField(BaseField):

    """
    Primitive that handles a binary input
    """

    def __init__(self, value: _BinaryInput):
        super().__init__(value)

    def validate(self, value: _BinaryInput | None = None) -> None:
        value = value or self.value
        if not is_binary_input(value):
            raise TypeError()

    @staticmethod
    def to_io(value: _BinaryInput) -> _IOOutput:
        if isinstance(value, bytes):
            io_obj = io.BytesIO(value)
            io_obj.name = uuid.uuid4().hex
        elif isinstance(value, (pathlib.Path, str)):
            io_obj = open(value, "rb")
        elif isinstance(value, io.BufferedIOBase):
            io_obj = value
        else:
            raise TypeError(
                f"Required one of: bytes, str, pathlib.Path, got {type(value)}"
            )
        return io_obj

    def serialize(
        self, value=None
    ) -> tuple[BinaryIO | BytesIO, str | Any, tuple[str | None, str | None] | str]:
        value = value or self.value
        io_obj = self.to_io(value)
        filename = io_obj.name
        mime = mimetypes.guess_type(filename)[0] or "applications/octet-stream"
        io_obj.seek(0)
        return filename, io_obj, mime



import httpx

from flymyai.core._response import FlyMyAIResponse
from flymyai.core._streaming import ServerSentEvent
from flymyai.core.exceptions import BaseFlyMyAIException


class ResponseFactoryException(Exception):
    ...


class ResponseFactory(object):
    """
    Factory for FlyMyAIResponse objects
    """

    def __init__(
        self,
        httpx_response: httpx.Response = None,
        httpx_request: httpx.Request = None,
        sse: ServerSentEvent = None,
    ):
        if not httpx_response and not sse:
            raise ResponseFactoryException("httpx_response and sse params required")
        if httpx_response and not httpx_request:
            self.httpx_request = httpx_response.request
        else:
            self.httpx_request = httpx_request
        self.sse = sse
        self.httpx_response = httpx_response

    def get_sse_status_code(self):
        return self.sse.json().get("status_code", 200)

    def _base_construct_from_sse(self):
        sse_status = self.get_sse_status_code()
        if sse_status < 400:
            return FlyMyAIResponse(
                status_code=sse_status,
                content=self.sse.data or self.sse.event,
                request=self.httpx_request,
                headers=self.httpx_response.headers or self.sse.headers,
            )
        else:
            raise BaseFlyMyAIException.from_response(
                FlyMyAIResponse(
                    status_code=sse_status,
                    content=self.sse.data or self.sse.event,
                    request=self.httpx_request,
                    headers=self.httpx_response.headers or self.sse.headers,
                )
            )

    def _base_construct_from_httpx_response(self):
        if self.httpx_response.status_code < 400:
            return FlyMyAIResponse.from_httpx(self.httpx_response)
        else:
            raise BaseFlyMyAIException.from_response(
                FlyMyAIResponse.from_httpx(self.httpx_response)
            )

    def construct(self):
        if self.sse:
            return self._base_construct_from_sse()
        elif self.httpx_response:
            return self._base_construct_from_httpx_response()



import dataclasses
import json

import httpx
import pydantic
from pydantic import PrivateAttr

from flymyai.core._response import FlyMyAIResponse


@dataclasses.dataclass
class Base4xxResponse:
    """
    Base class for all 4xx
    """

    status_code: int
    url: httpx.URL
    content: bytes

    requires_retry: bool = False

    def to_msg(self):
        return f"""
            BAD REQUEST DETECTED ({self.status_code}):
            REQUEST URL: {self.url};
        """

    @classmethod
    def from_response(cls, response: httpx.Response):
        return cls(response.status_code, response.url, response.content)


@dataclasses.dataclass
class FlyMyAI400Response(Base4xxResponse):
    """
    400 response
    requires_retry = False means we do not resend this request
    """

    requires_retry: bool = False

    def to_msg(self):
        return f"""
            Bad request happened: {self.content.decode()}
        """


@dataclasses.dataclass
class FlyMyAI401Response(Base4xxResponse):
    """
    401 response
    requires_retry = False means we do not resend this request
    """

    requires_retry: bool = False

    def to_msg(self):
        return f"""
            Authentication error: verify your credentials!
        """


@dataclasses.dataclass
class FlyMyAI422Response(Base4xxResponse):
    """
    422 response
    requires_retry = False means we do not resend this request
    """

    requires_retry: bool = False

    def to_msg(self):
        jsoned = json.loads(self.content)
        msg = super().to_msg()
        if detail := jsoned.get("detail"):
            msg += f"Details: {detail}"
        return msg


class PredictionResponse(pydantic.BaseModel):
    """
    Prediction response from FlyMyAI
    """

    exc_history: list | None
    output_data: dict
    _response: FlyMyAIResponse = PrivateAttr()

    inference_time: float | None = None

    def __init__(self, **data):
        super().__init__(**data)
        self._response = data.get("response")

    @property
    def response(self):
        return self._response


class OpenAPISchemaResponse(pydantic.BaseModel):
    """
    OpenAPI schema for current project. Use it to construct your own schema
    """

    exc_history: list | None
    openapi_schema: dict
    _response: FlyMyAIResponse = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._response = data.get("response")

    @property
    def response(self):
        return self._response
